// Coherent noise terrain generation interface implementation. Makes use of NoiseLib

#include "CoherentNoise.h"
#include "noiseutils.h"

using namespace noise;
using namespace std;

//class connstructor
CoherentNoiseTerrain::CoherentNoiseTerrain(int size)
{
	miSize = size;
}

//class destructor implementation
CoherentNoiseTerrain::~CoherentNoiseTerrain()
{
	//implement me
}

void CoherentNoiseTerrain::MultiFractalPerlin(double octaves, double baseFrequency, double frequency, double persistance, double lowerX, double upperX, double lowerZ , double upperZ)
{
	srand(time(NULL));
	randSeed = rand();
	mountainTerrain.SetSeed(randSeed);
	baseFlatTerrain.SetSeed(randSeed);
	terrainType.SetSeed(randSeed);

	baseFlatTerrain.SetFrequency (baseFrequency);

	flatTerrain.SetSourceModule (0, baseFlatTerrain);
	flatTerrain.SetScale (0.8);
	flatTerrain.SetBias (15.0);

	terrainType.SetOctaveCount (octaves); //suggested value: 6
	terrainType.SetFrequency (frequency); //suggested value: 0.5
	terrainType.SetPersistence (persistance); //suggested value: 0.25

	finalTerrain.SetSourceModule (0, flatTerrain);
	finalTerrain.SetSourceModule (1, mountainTerrain);
	finalTerrain.SetControlModule (terrainType);
	finalTerrain.SetBounds (0.0, 1000.0);
	finalTerrain.SetEdgeFalloff (0.125);

	heightMapBuilder.SetSourceModule (finalTerrain);
	heightMapBuilder.SetDestNoiseMap (heightMap);
	heightMapBuilder.SetDestSize (miSize, miSize);
	heightMapBuilder.SetBounds (lowerX, upperX, lowerZ, upperZ); //suggested values: 6.0, 10.0, 1.0, 5.0
	heightMapBuilder.Build ();

	for(int i = 0; i < miSize ; ++i)
	{
		for(int j = 0; j < miSize; ++j)
		{
			mfField[(j * miSize) + i] = heightMap.GetValue(i, j);
		}
	}

	//Generate();
}


//Used debug stuff only atm...
void CoherentNoiseTerrain::Generate()
{
	//call any of the generators here, if needed

	renderer.SetSourceNoiseMap (heightMap);
	renderer.SetDestImage (image);
	renderer.ClearGradient ();
	renderer.AddGradientPoint ( 15.00, utils::Color ( 32, 160,   0, 255)); // grass
	renderer.AddGradientPoint ( 2.25, utils::Color (224, 224,   0, 255)); // dirt
	renderer.AddGradientPoint ( 1.25, utils::Color (128, 128, 128, 255)); // rock
	renderer.AddGradientPoint ( -15.0, utils::Color (0, 0, 255, 255)); // water
	renderer.Render ();

	utils::WriterBMP writer;
	writer.SetSourceImage (image);
	writer.SetDestFilename ("../data/terrainTex.bmp");
	writer.WriteDestFile ();
}
