/////////////////////////////////////////////////////
// Auxilary maths interface.
// Version: 020412_01
////////////////////////////////////////////////////

#ifndef _AUXMATHS_H_
#define _AUXMATHS_H_

#include "vector3.h"

#define PI 3.14159265

inline float Lerp(float t, float a, float b)
{
	return (a + t * (b - a));
}

/////////////////////////////////////////////////////
//Line/Trianlge interesction checking routines.
/////////////////////////////////////////////////////

inline float PlaneDistance(Vector3 Normal, Vector3 Point)
{	
	float distance = 0;			

	distance = - ((Normal.x * Point.x) + (Normal.y * Point.y) + (Normal.z * Point.z));

	return distance;			
}

inline bool IntersectedPlane(Vector3 vPoly[], Vector3 vLine[], Vector3 &vNormal, float &originDistance)
{
	float distance1=0, distance2=0;						
			
	vNormal = vNormal.Normal(vPoly);	

	originDistance = PlaneDistance(vNormal, vPoly[0]);

	distance1 = ((vNormal.x * vLine[0].x)  +					// Ax +
		         (vNormal.y * vLine[0].y)  +					// Bx +
				 (vNormal.z * vLine[0].z)) + originDistance;	// Cz + D
	
	distance2 = ((vNormal.x * vLine[1].x)  +					// Ax +
		         (vNormal.y * vLine[1].y)  +					// Bx +
				 (vNormal.z * vLine[1].z)) + originDistance;	// Cz + D


	if(distance1 * distance2 >= 0)			
	   return false;					
					
	return true;							
}

inline Vector3 IntersectionPoint(Vector3 vNormal, Vector3 vLine[], double distance)
{
	Vector3 vPoint; 
	Vector3 vLineDir;		
	double Numerator = 0.0, Denominator = 0.0, dist = 0.0;

	vLineDir = vLineDir.Vector(vLine[1], vLine[0]);		
	vLineDir = vLineDir.Normalize(vLineDir);				

	Numerator = - (vNormal.x * vLine[0].x +		
				   vNormal.y * vLine[0].y +
				   vNormal.z * vLine[0].z + distance);

	Denominator = vNormal.Dot(vNormal, vLineDir);		
				  
	if( Denominator == 0.0)			
	{
		return vLine[0];	
	}

	dist = Numerator / Denominator;				
	
	vPoint.x = (float)(vLine[0].x + (vLineDir.x * dist));
	vPoint.y = (float)(vLine[0].y + (vLineDir.y * dist));
	vPoint.z = (float)(vLine[0].z + (vLineDir.z * dist));

	return vPoint;								
}

inline bool InsidePolygon(Vector3 vIntersection, Vector3 Poly[], long verticeCount)
{
	const double MATCH_FACTOR = 0.9999;		
	double Angle = 0.0;						
	Vector3 vA, vB;						

	for (int i = 0; i < verticeCount; i++)		
	{	
		vA = vA.Vector(Poly[i], vIntersection);	
												
		vB = vB.Vector(Poly[(i + 1) % verticeCount], vIntersection);
												
		Angle += vA.AngleBetweenVectors(vA, vB);	
	}
												
	if(Angle >= (MATCH_FACTOR * (2.0 * PI)) )	
		return true;							
		
	return false;							
}

inline bool IntersectedPolygon(Vector3 vPoly[], Vector3 vLine[], int verticeCount)
{
	Vector3 vNormal;
	float originDistance = 0;

	if(!IntersectedPlane(vPoly, vLine,   vNormal,   originDistance))
	{
		return false;
	}

	Vector3 vIntersection = IntersectionPoint(vNormal, vLine, originDistance);

	if(InsidePolygon(vIntersection, vPoly, verticeCount))
	{
		return true;
	}

	return false;								
}

#endif