////////////////////////////////////////////////////////////////////////////
// New chunk interface implementation.
// Written by Vladeta Stojanovic. Based on the Glescraft tutorial code.
//
// Version: 180412_01
////////////////////////////////////////////////////////////////////////////

//todo: Make the root node of the octree be th eusers current camera position, the expand the lead nodes into the mseh

#include "chunk.h"
#include "cubeTerrain.h"
#include "frustum.h"
#include "auxMaths.h"
#include <Windows.h>
#include <iostream>
#include <math.h>

using namespace std;

//default green
static const GLfloat ambientMat0 [] = {0.1, 0.1, 0.1, 1.00}; 
static const GLfloat diffuseMat0 [] = {0.5, 1.0, 0.5, 1.00}; 
static const GLfloat specMat0 [] = {0.00, 0.50, 0.00, 0.50}; 
static bool drawWire = false;

CoherentNoiseTerrain gChunkNoise(32);
CoherentNoiseTerrain gChunkNoiseT(32);
Frustum gChunkFrustum;

bool detectcollision(int x1, int y1, int z1,int radis1, int x2, int y2, int z2, int radis2)
{
    int xd, yd, zd, Distance;

    xd = x2-x1;
    yd = y2-y1;
    zd = z2-z1;
    Distance = (int)sqrtf(xd*xd + yd*yd + zd*zd);
  
     
    if( radis1 + radis2 >= Distance)
        return true; //collision
    return false;    //no collision   
}

Chunk::Chunk()
{
	gChunkNoise.MultiFractalPerlin(10, 2, 0.4, 0.25, 10, 20, 7, 25);
	gChunkNoiseT.MultiFractalPerlin(10, 2, 0.4, 0.25, 10, 20, 7, 25);

	vboID = NULL;
	indexVBOID = NULL;

	mVolumeSize = gChunkNoise.ReturnSize();
	mHeightLimit = 14.0f;
	mVolumeDepth = 32.0f;

	chunkX = new float[mVolumeSize * mVolumeSize * mVolumeDepth];
	chunkY = new float[mVolumeSize * mVolumeSize * mVolumeDepth];
	chunkZ = new float[mVolumeSize * mVolumeSize * mVolumeDepth];
	mMaxY = new float[mVolumeSize * mVolumeSize * mVolumeDepth];
	chkInfo = new ChunkInfo[mVolumeSize * mVolumeSize * mVolumeDepth];

	for(int i = 0; i < (mVolumeSize * mVolumeSize * mVolumeDepth); ++i)
	{
		chunkX[i] = 0;
		chunkY[i] = 0;
		chunkZ[i] = 0;
		mMaxY[i] = 0;
	}
}

Chunk::~Chunk()
{
  	free(chunkX);
	free(chunkY);
	free(chunkZ);
	free(mMaxY);
}

void Chunk::ReInit(int size)
{
	free(chunkX);
	free(chunkY);
	free(chunkZ);
	free(mMaxY);

	gChunkNoise.ResetSize(size);
	gChunkNoise.MultiFractalPerlin(10, 2, 0.4, 0.25, 10, 20, 7, 25);

	chunkX = new float[mVolumeSize * mVolumeSize * mVolumeDepth];
	chunkY = new float[mVolumeSize * mVolumeSize * mVolumeDepth];
	chunkZ = new float[mVolumeSize * mVolumeSize * mVolumeDepth];
	mMaxY = new float[mVolumeSize * mVolumeSize * mVolumeDepth];

	for(int i = 0; i < (mVolumeSize * mVolumeSize * mVolumeDepth); ++i)
	{
		chunkX[i] = 0;
		chunkY[i] = 0;
		chunkZ[i] = 0;
		mMaxY[i] = 0;
	}
}

void Chunk::InitVBO()
{
	glGenBuffers(1, &vboID);
	glBindBuffer(GL_ARRAY_BUFFER, vboID); 
	
	glBufferData(GL_ARRAY_BUFFER, sizeof(Vertex) * 24, NULL, GL_STATIC_DRAW);
	glBufferSubData(GL_ARRAY_BUFFER, 0, sizeof(Vertex) * 24, verts);
	
	glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), BUFFER_OFFSET(12));
	glNormalPointer(GL_FLOAT, sizeof(Vertex), BUFFER_OFFSET(20));
	glColorPointer(4, GL_FLOAT, sizeof(Vertex), BUFFER_OFFSET(32));
	glVertexPointer(3, GL_FLOAT, sizeof(Vertex), BUFFER_OFFSET(0));
	
	glGenBuffers(1, &indexVBOID); 
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexVBOID); 
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, 36 * sizeof(GLubyte), index, GL_STATIC_DRAW);
}

void Chunk::BuildCube()
{
	//Top
	verts[0].location[X_POS]	= 1;	verts[0].location[Y_POS]	= 1;	verts[0].location[Z_POS]	= -1;
	verts[0].normal[X_POS]		= 0;	verts[0].normal[Y_POS]		= 1;	verts[0].normal[Z_POS]		= 0;
	verts[0].tex[U_POS]			= 0;	verts[0].tex[V_POS]			= 0; 
	
	verts[1].location[X_POS]	= -1;	verts[1].location[Y_POS]	= 1;	verts[1].location[Z_POS]	= -1;
	verts[1].normal[X_POS]		= 0;	verts[1].normal[Y_POS]		= 1;	verts[1].normal[Z_POS]		= 0;
	verts[1].tex[U_POS]			= 0;	verts[1].tex[V_POS]			= 1; 
	
	verts[2].location[X_POS]	= -1;	verts[2].location[Y_POS]	= 1;	verts[2].location[Z_POS]	= 1;
	verts[2].normal[X_POS]		= 0;	verts[2].normal[Y_POS]		= 1;	verts[2].normal[Z_POS]		= 0;
	verts[2].tex[U_POS]			= 1;	verts[2].tex[V_POS]			= 1; 
	
	verts[3].location[X_POS]	= 1;	verts[3].location[Y_POS]	= 1;	verts[3].location[Z_POS]	= 1;
	verts[3].normal[X_POS]		= 0;	verts[3].normal[Y_POS]		= 1;	verts[3].normal[Z_POS]		= 0;
	verts[3].tex[U_POS]			= 1;	verts[3].tex[V_POS]			= 0;
	
	// Bottom
	verts[4].location[X_POS]	= 1;	verts[4].location[Y_POS]	= -1;	verts[4].location[Z_POS]	= 1;
	verts[4].normal[X_POS]		= 0;	verts[4].normal[Y_POS]		= -1;	verts[4].normal[Z_POS]		= 0;
	verts[4].tex[U_POS]			= 0;	verts[4].tex[V_POS]			= 0; 
	
	verts[5].location[X_POS]	= -1;	verts[5].location[Y_POS]	= -1;	verts[5].location[Z_POS]	= 1;
	verts[5].normal[X_POS]		= 0;	verts[5].normal[Y_POS]		= -1;	verts[5].normal[Z_POS]		= 0;
	verts[5].tex[U_POS]			= 0;	verts[5].tex[V_POS]			= 1; 
	
	verts[6].location[X_POS]	= -1;	verts[6].location[Y_POS]	= -1;	verts[6].location[Z_POS]	= -1;
	verts[6].normal[X_POS]		= 0;	verts[6].normal[Y_POS]		= -1;	verts[6].normal[Z_POS]		= 0;
	verts[6].tex[U_POS]			= 1;	verts[6].tex[V_POS]			= 1; 
	
	verts[7].location[X_POS]	= 1;	verts[7].location[Y_POS]	= -1;	verts[7].location[Z_POS]	= -1;
	verts[7].normal[X_POS]		= 0;	verts[7].normal[Y_POS]		= -1;	verts[7].normal[Z_POS]		= 0;
	verts[7].tex[U_POS]			= 1;	verts[7].tex[V_POS]			= 0;
	
	// Front
	verts[8].location[X_POS]	= 1;	verts[8].location[Y_POS]	= 1;	verts[8].location[Z_POS]	= 1;
	verts[8].normal[X_POS]		= 0;	verts[8].normal[Y_POS]		= 0;	verts[8].normal[Z_POS]		= 1;
	verts[8].tex[U_POS]			= 0;	verts[8].tex[V_POS]			= 0; 
	
	verts[9].location[X_POS]	= -1;	verts[9].location[Y_POS]	= 1;	verts[9].location[Z_POS]	= 1;
	verts[9].normal[X_POS]		= 0;	verts[9].normal[Y_POS]		= 0;	verts[9].normal[Z_POS]		= 1;
	verts[9].tex[U_POS]			= 0;	verts[9].tex[V_POS]			= 1; 
	
	verts[10].location[X_POS]	= -1;	verts[10].location[Y_POS]	= -1;	verts[10].location[Z_POS]	= 1;
	verts[10].normal[X_POS]		= 0;	verts[10].normal[Y_POS]		= 0;	verts[10].normal[Z_POS]		= 1;
	verts[10].tex[U_POS]		= 1;	verts[10].tex[V_POS]		= 1; 
	
	verts[11].location[X_POS]	= 1;	verts[11].location[Y_POS]	= -1;	verts[11].location[Z_POS]	= 1;
	verts[11].normal[X_POS]		= 0;	verts[11].normal[Y_POS]		= 0;	verts[11].normal[Z_POS]		= 1;
	verts[11].tex[U_POS]		= 1;	verts[11].tex[V_POS]		= 0;

	// Back
	verts[12].location[X_POS]	= 1;	verts[12].location[Y_POS]	= -1;	verts[12].location[Z_POS]	= -1;
	verts[12].normal[X_POS]		= 0;	verts[12].normal[Y_POS]		= 0;	verts[12].normal[Z_POS]		= -1;
	verts[12].tex[U_POS]		= 0;	verts[12].tex[V_POS]		= 0; 
	
	verts[13].location[X_POS]	= -1;	verts[13].location[Y_POS]	= -1;	verts[13].location[Z_POS]	= -1;
	verts[13].normal[X_POS]		= 0;	verts[13].normal[Y_POS]		= 0;	verts[13].normal[Z_POS]		= -1;
	verts[13].tex[U_POS]		= 0;	verts[13].tex[V_POS]		= 1; 
	
	verts[14].location[X_POS]	= -1;	verts[14].location[Y_POS]	= 1;	verts[14].location[Z_POS]	= -1;
	verts[14].normal[X_POS]		= 0;	verts[14].normal[Y_POS]		= 0;	verts[14].normal[Z_POS]		= -1;
	verts[14].tex[U_POS]		= 1;	verts[14].tex[V_POS]		= 1; 
	
	verts[15].location[X_POS]	= 1;	verts[15].location[Y_POS]	= 1;	verts[15].location[Z_POS]	= -1;
	verts[15].normal[X_POS]		= 0;	verts[15].normal[Y_POS]		= 0;	verts[15].normal[Z_POS]		= -1;
	verts[15].tex[U_POS]		= 1;	verts[15].tex[V_POS]		= 0;

	// Left
	verts[16].location[X_POS]	= -1;	verts[16].location[Y_POS]	= 1;	verts[16].location[Z_POS]	= 1;
	verts[16].normal[X_POS]		= -1;	verts[16].normal[Y_POS]		= 0;	verts[16].normal[Z_POS]		= 0;
	verts[16].tex[U_POS]		= 0;	verts[16].tex[V_POS]		= 0; 
	
	verts[17].location[X_POS]	= -1;	verts[17].location[Y_POS]	= 1;	verts[17].location[Z_POS]	= -1;
	verts[17].normal[X_POS]		= -1;	verts[17].normal[Y_POS]		= 0;	verts[17].normal[Z_POS]		= 0;
	verts[17].tex[U_POS]		= 0;	verts[17].tex[V_POS]		= 1; 
	
	verts[18].location[X_POS]	= -1;	verts[18].location[Y_POS]	= -1;	verts[18].location[Z_POS]	= -1;
	verts[18].normal[X_POS]		= -1;	verts[18].normal[Y_POS]		= 0;	verts[18].normal[Z_POS]		= 0;
	verts[18].tex[U_POS]		= 1;	verts[18].tex[V_POS]		= 1; 
	
	verts[19].location[X_POS]	= -1;	verts[19].location[Y_POS]	= -1;	verts[19].location[Z_POS]	= 1;
	verts[19].normal[X_POS]		= -1;	verts[19].normal[Y_POS]		= 0;	verts[19].normal[Z_POS]		= 0;
	verts[19].tex[U_POS]		= 1;	verts[19].tex[V_POS]		= 0;

	// Right
	verts[20].location[X_POS]	= 1;	verts[20].location[Y_POS]	= 1;	verts[20].location[Z_POS]	= -1;
	verts[20].normal[X_POS]		= 1;	verts[20].normal[Y_POS]		= 0;	verts[20].normal[Z_POS]		= 0;
	verts[20].tex[U_POS]		= 0;	verts[20].tex[V_POS]		= 0; 
	
	verts[21].location[X_POS]	= 1;	verts[21].location[Y_POS]	= 1;	verts[21].location[Z_POS]	= 1;
	verts[21].normal[X_POS]		= 1;	verts[21].normal[Y_POS]		= 0;	verts[21].normal[Z_POS]		= 0;
	verts[21].tex[U_POS]		= 0;	verts[21].tex[V_POS]		= 1; 
	
	verts[22].location[X_POS]	= 1;	verts[22].location[Y_POS]	= -1;	verts[22].location[Z_POS]	= 1;
	verts[22].normal[X_POS]		= 1;	verts[22].normal[Y_POS]		= 0;	verts[22].normal[Z_POS]		= 0;
	verts[22].tex[U_POS]		= 1;	verts[22].tex[V_POS]		= 1; 
	
	verts[23].location[X_POS]	= 1;	verts[23].location[Y_POS]	= -1;	verts[23].location[Z_POS]	= -1;
	verts[23].normal[X_POS]		= 1;	verts[23].normal[Y_POS]		= 0;	verts[23].normal[Z_POS]		= 0;
	verts[23].tex[U_POS]		= 1;	verts[23].tex[V_POS]		= 0;
	
	// Colours
	for (int i = 0; i < 24; i++)
	{
		verts[i].colour[R_POS] = 1.0;
		verts[i].colour[G_POS] = 1.0;
		verts[i].colour[B_POS] = 1.0;
		verts[i].colour[A_POS] = 1.0;
	}
	
	index[0] = 0;	index[1] = 1;	index[2] = 2;
	index[3] = 2;	index[4] = 3;	index[5] = 0;
	
	index[6] = 4;	index[7] = 5;	index[8] = 6;
	index[9] = 6;	index[10] = 7;	index[11] = 4;
	
	index[12] = 8;	index[13] = 9;	index[14] = 10;
	index[15] = 10;	index[16] = 11;	index[17] = 8;
	
	index[18] = 12;	index[19] = 13;	index[20] = 14;
	index[21] = 14;	index[22] = 15;	index[23] = 12;
	
	index[24] = 16;	index[25] = 17;	index[26] = 18;
	index[27] = 18;	index[28] = 19;	index[29] = 16;
	
	index[30] = 20;	index[31] = 21;	index[32] = 22;
	index[33] = 22;	index[34] = 23;	index[35] = 20;
}

void Chunk::DrawChunks(float x, float y, float z, int step, Camera* cam)
{
	glPushMatrix();

	chunkX[step] = x;
	chunkY[step] = y;
	chunkZ[step] = z;

	glTranslatef(chunkX[step], chunkY[step], chunkZ[step]);

	if(chunkY[step] > mHeightLimit) 
	{
		if(detectcollision(chunkX[step], chunkY[step], chunkZ[step], 100, cam->GetPosX(), cam->GetPosY(), cam->GetPosZ(), 100))
		{
			glDrawElements(GL_TRIANGLES, 36, GL_UNSIGNED_BYTE, BUFFER_OFFSET(0));
		}
		else
		{
			if(GetKeyState('V') & 0x80)
				glDrawElements(GL_LINES, 36, GL_UNSIGNED_BYTE, BUFFER_OFFSET(0));
		}
	}

	glPopMatrix();
  
}

void Chunk::CalculateVisibility(int i, int j, int step)
{
	chkInfo[step].id = step;
	chkInfo[step].pos.x = i;
	chkInfo[step].pos.y = step;
	chkInfo[step].pos.z = j;
	chkInfo[step].visible = true;
	chkInfo[step].active = true;
}

void Chunk::RenderVBOTerrain(float scaleX, float scaleY, float scaleZ, int depth, Camera* cam)
{
	mVolumeDepth = depth;

	gChunkFrustum.ExtractFrustum();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glBindBuffer(GL_ARRAY_BUFFER, vboID);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexVBOID);
	
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);
	
	glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), BUFFER_OFFSET(12));
	glNormalPointer(GL_FLOAT, sizeof(Vertex), BUFFER_OFFSET(20));
	glColorPointer(4, GL_FLOAT, sizeof(Vertex), BUFFER_OFFSET(32));
	glVertexPointer(3, GL_FLOAT, sizeof(Vertex), BUFFER_OFFSET(0));

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,   ambientMat0); 
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,   diffuseMat0); 
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,  specMat0); 
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 0.5f);
	
	glEnable(GL_DEPTH_TEST);

	glPushMatrix();

		glScalef(scaleX, scaleY, scaleZ);
		for(int i = 0; i < gChunkNoise.ReturnSize() ; ++i)
		{
			for(int j = 0; j < gChunkNoise.ReturnSize(); ++j)
			{
				for(int k = 0; k <  mVolumeDepth; ++k)
				{
					//pass in all the chunk info here for sorting (not currently used)
					CalculateVisibility(i, j, k + (j * gChunkNoise.ReturnSize() + i));

					if(gChunkFrustum.CheckCubeFrustum(chunkX[k + (j * gChunkNoise.ReturnSize() + i)], chunkY[k + (j * gChunkNoise.ReturnSize() + i)], chunkZ[k + (j * gChunkNoise.ReturnSize() + i)], 4))
					{
						if(chkInfo[k + (j * gChunkNoise.ReturnSize() + i)].visible)
						DrawChunks(i, k + gChunkNoise.mfField[(j * gChunkNoise.ReturnSize()) + i], j, k + (j * gChunkNoise.ReturnSize() + i), cam);
					}
				}
			}
		}

	glPopMatrix();

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisable(GL_DEPTH_TEST);

}