////////////////////////////////////////////////////////////////////////////
// New chunk interface.
// Written by Vladeta Stojanovic. Based on the Glescraft tutorial code.
//
// Version: 250312_01
////////////////////////////////////////////////////////////////////////////

#ifndef _CHUNK_H_
#define _CHUNK_H_

#include <GL/glew.h>
#include <GL/freeglut.h>
#include <stdio.h>
#include <stdlib.h>
#include "auxMaths.h"
#include "vector3.h"
#include "camera.h"

#define BUFFER_OFFSET(i) ((char *)NULL + (i))

// Location/Normals
#define X_POS 0
#define Y_POS 1
#define Z_POS 2

// Texture Coordinates
#define U_POS 0
#define V_POS 1

// Colours
#define R_POS 0
#define G_POS 1
#define B_POS 2
#define A_POS 3

typedef struct
{
	GLfloat location[3];
	GLfloat tex[2];
	GLfloat normal[3];
	GLfloat colour[4];
	GLubyte padding[16];
} Vertex;

typedef struct
{
	int id;
	bool visible;
	bool active;
	Vector3 pos;
} ChunkInfo;

class Chunk
{
public:

	Chunk();
	~Chunk();

	void InitVBO();
	void BuildCube();
	void DrawChunks(float x, float y, float z, int step, Camera* cam);
	void CalculateVisibility(int i, int j, int step);
	void RenderVBOTerrain(float scaleX, float scaleY, float scaleZ, int depth , Camera* cam);
	void ReInit(int size);
	

public:

	int		mVolumeSize;
	int		mVolumeDepth;
	float	mHeightLimit;

private:

	Vertex  verts[24];
	GLubyte index[36];
	GLuint  vboID;
	GLuint  indexVBOID;

public:
	float*	chunkX;
	float*	chunkY;
	float*	chunkZ;
	float*	mMaxY;
	ChunkInfo* chkInfo;
};

extern Chunk gChunk;



#endif