////////////////////////////////////////////////
// Frustum culling interface implementation.
//
// Version: 170112_01
////////////////////////////////////////////////

#include "frustum.h"

enum PlaneData
{
	A = 0,				// The X value of the plane's normal
	B = 1,				// The Y value of the plane's normal
	C = 2,				// The Z value of the plane's normal
	D = 3				// The distance the plane is from the origin
};

Frustum::Frustum()
{
	//implement me
}

Frustum::~Frustum()
{
	//implement me
}

void Frustum::ExtractFrustum()
{
	//get current projection matrix
	glGetFloatv(GL_PROJECTION_MATRIX, mProj);

	//get current model matrix
	glGetFloatv(GL_MODELVIEW_MATRIX, mModel);

	//Set the clipping matrix by multiplying the projection matrix by the view matrix
	mClip[ 0] = mModel[ 0] * mProj[ 0] + mModel[ 1] * mProj[ 4] + mModel[ 2] * mProj[ 8] + mModel[ 3] * mProj[12];
	mClip[ 1] = mModel[ 0] * mProj[ 1] + mModel[ 1] * mProj[ 5] + mModel[ 2] * mProj[ 9] + mModel[ 3] * mProj[13];
	mClip[ 2] = mModel[ 0] * mProj[ 2] + mModel[ 1] * mProj[ 6] + mModel[ 2] * mProj[10] + mModel[ 3] * mProj[14];
	mClip[ 3] = mModel[ 0] * mProj[ 3] + mModel[ 1] * mProj[ 7] + mModel[ 2] * mProj[11] + mModel[ 3] * mProj[15];

	mClip[ 4] = mModel[ 4] * mProj[ 0] + mModel[ 5] * mProj[ 4] + mModel[ 6] * mProj[ 8] + mModel[ 7] * mProj[12];
	mClip[ 5] = mModel[ 4] * mProj[ 1] + mModel[ 5] * mProj[ 5] + mModel[ 6] * mProj[ 9] + mModel[ 7] * mProj[13];
	mClip[ 6] = mModel[ 4] * mProj[ 2] + mModel[ 5] * mProj[ 6] + mModel[ 6] * mProj[10] + mModel[ 7] * mProj[14];
	mClip[ 7] = mModel[ 4] * mProj[ 3] + mModel[ 5] * mProj[ 7] + mModel[ 6] * mProj[11] + mModel[ 7] * mProj[15];

	mClip[ 8] = mModel[ 8] * mProj[ 0] + mModel[ 9] * mProj[ 4] + mModel[10] * mProj[ 8] + mModel[11] * mProj[12];
	mClip[ 9] = mModel[ 8] * mProj[ 1] + mModel[ 9] * mProj[ 5] + mModel[10] * mProj[ 9] + mModel[11] * mProj[13];
	mClip[10] = mModel[ 8] * mProj[ 2] + mModel[ 9] * mProj[ 6] + mModel[10] * mProj[10] + mModel[11] * mProj[14];
	mClip[11] = mModel[ 8] * mProj[ 3] + mModel[ 9] * mProj[ 7] + mModel[10] * mProj[11] + mModel[11] * mProj[15];

	mClip[12] = mModel[12] * mProj[ 0] + mModel[13] * mProj[ 4] + mModel[14] * mProj[ 8] + mModel[15] * mProj[12];
	mClip[13] = mModel[12] * mProj[ 1] + mModel[13] * mProj[ 5] + mModel[14] * mProj[ 9] + mModel[15] * mProj[13];
	mClip[14] = mModel[12] * mProj[ 2] + mModel[13] * mProj[ 6] + mModel[14] * mProj[10] + mModel[15] * mProj[14];
	mClip[15] = mModel[12] * mProj[ 3] + mModel[13] * mProj[ 7] + mModel[14] * mProj[11] + mModel[15] * mProj[15];

	/////////////////////////////////////
	//Frustum extraction calculations
	/////////////////////////////////////

   //Extract the numbers for the RIGHT plane 
   mFrustum[0][0] = mClip[ 3] - mClip[ 0];
   mFrustum[0][1] = mClip[ 7] - mClip[ 4];
   mFrustum[0][2] = mClip[11] - mClip[ 8];
   mFrustum[0][3] = mClip[15] - mClip[12];

   //Normalize the result 
   t = sqrt( mFrustum[0][0] * mFrustum[0][0] + mFrustum[0][1] * mFrustum[0][1] + mFrustum[0][2] * mFrustum[0][2] );
   mFrustum[0][0] /= t;
   mFrustum[0][1] /= t;
   mFrustum[0][2] /= t;
   mFrustum[0][3] /= t;

   //Extract the numbers for the LEFT plane 
   mFrustum[1][0] = mClip[ 3] + mClip[ 0];
   mFrustum[1][1] = mClip[ 7] + mClip[ 4];
   mFrustum[1][2] = mClip[11] + mClip[ 8];
   mFrustum[1][3] = mClip[15] + mClip[12];

   //Normalize the result 
   t = sqrt( mFrustum[1][0] * mFrustum[1][0] + mFrustum[1][1] * mFrustum[1][1] + mFrustum[1][2] * mFrustum[1][2] );
   mFrustum[1][0] /= t;
   mFrustum[1][1] /= t;
   mFrustum[1][2] /= t;
   mFrustum[1][3] /= t;

   //Extract the BOTTOM plane 
   mFrustum[2][0] = mClip[ 3] + mClip[ 1];
   mFrustum[2][1] = mClip[ 7] + mClip[ 5];
   mFrustum[2][2] = mClip[11] + mClip[ 9];
   mFrustum[2][3] = mClip[15] + mClip[13];

   //Normalize the result 
   t = sqrt( mFrustum[2][0] * mFrustum[2][0] + mFrustum[2][1] * mFrustum[2][1] + mFrustum[2][2] * mFrustum[2][2] );
   mFrustum[2][0] /= t;
   mFrustum[2][1] /= t;
   mFrustum[2][2] /= t;
   mFrustum[2][3] /= t;

   //Extract the TOP plane 
   mFrustum[3][0] = mClip[ 3] - mClip[ 1];
   mFrustum[3][1] = mClip[ 7] - mClip[ 5];
   mFrustum[3][2] = mClip[11] - mClip[ 9];
   mFrustum[3][3] = mClip[15] - mClip[13];

   //Normalize the result 
   t = sqrt( mFrustum[3][0] * mFrustum[3][0] + mFrustum[3][1] * mFrustum[3][1] + mFrustum[3][2] * mFrustum[3][2] );
   mFrustum[3][0] /= t;
   mFrustum[3][1] /= t;
   mFrustum[3][2] /= t;
   mFrustum[3][3] /= t;

   //Extract the FAR plane 
   mFrustum[4][0] = mClip[ 3] - mClip[ 2];
   mFrustum[4][1] = mClip[ 7] - mClip[ 6];
   mFrustum[4][2] = mClip[11] - mClip[10];
   mFrustum[4][3] = mClip[15] - mClip[14];

   //Normalize the result 
   t = sqrt( mFrustum[4][0] * mFrustum[4][0] + mFrustum[4][1] * mFrustum[4][1] + mFrustum[4][2] * mFrustum[4][2] );
   mFrustum[4][0] /= t;
   mFrustum[4][1] /= t;
   mFrustum[4][2] /= t;
   mFrustum[4][3] /= t;

   //Extract the NEAR plane 
   mFrustum[5][0] = mClip[ 3] + mClip[ 2];
   mFrustum[5][1] = mClip[ 7] + mClip[ 6];
   mFrustum[5][2] = mClip[11] + mClip[10];
   mFrustum[5][3] = mClip[15] + mClip[14];

   //Normalize the result 
   t = sqrt( mFrustum[5][0] * mFrustum[5][0] + mFrustum[5][1] * mFrustum[5][1] + mFrustum[5][2] * mFrustum[5][2] );
   mFrustum[5][0] /= t;
   mFrustum[5][1] /= t;
   mFrustum[5][2] /= t;
   mFrustum[5][3] /= t;
}

bool Frustum::CheckCubeFrustum(float x, float y, float z, float size)
{
   int p;

   for( p = 0; p < 6; p++ )
   {
      if( mFrustum[p][0] * (x - size) + mFrustum[p][1] * (y - size) + mFrustum[p][2] * (z - size) + mFrustum[p][3] > 0 )
         continue;
      if( mFrustum[p][0] * (x + size) + mFrustum[p][1] * (y - size) + mFrustum[p][2] * (z - size) + mFrustum[p][3] > 0 )
         continue;
      if( mFrustum[p][0] * (x - size) + mFrustum[p][1] * (y + size) + mFrustum[p][2] * (z - size) + mFrustum[p][3] > 0 )
         continue;
      if( mFrustum[p][0] * (x + size) + mFrustum[p][1] * (y + size) + mFrustum[p][2] * (z - size) + mFrustum[p][3] > 0 )
         continue;
      if( mFrustum[p][0] * (x - size) + mFrustum[p][1] * (y - size) + mFrustum[p][2] * (z + size) + mFrustum[p][3] > 0 )
         continue;
      if( mFrustum[p][0] * (x + size) + mFrustum[p][1] * (y - size) + mFrustum[p][2] * (z + size) + mFrustum[p][3] > 0 )
         continue;
      if( mFrustum[p][0] * (x - size) + mFrustum[p][1] * (y + size) + mFrustum[p][2] * (z + size) + mFrustum[p][3] > 0 )
         continue;
      if( mFrustum[p][0] * (x + size) + mFrustum[p][1] * (y + size) + mFrustum[p][2] * (z + size) + mFrustum[p][3] > 0 )
         continue;
      return false;
   }
   return true;
}
