////////////////////////////////////////////////////////////////////////////
// OpenGL 2.0 Voxel terrain rendering demo.
// Written by Vladeta Stojanovic (vladeta_stojanovic@yahoo.co.uk).
//
// Version: 120512_01
////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#undef ERROR
#include <stdio.h>
#include <stdlib.h>
#include <GL/glew.h>
#include <GL/freeglut.h>
#include <iostream>
#include "textFile.h"
#include "camera.h"
#include "auxMaths.h"
#include "chunk.h"

using namespace std;

#define printOpenGLError() printOglError(__FILE__, __LINE__)

///////////////////////////////////////////////////
//Global application object instance variables
///////////////////////////////////////////////////
Camera *gCamera; //camera
Chunk   gChunk;

bool lerpTerrain = false;
bool useShaders = true;

//FPS stuff
static unsigned int fps_start = 0;
static unsigned int fps_frames = 0;
char fpsBuffer[33];
char sizeBuffer[8];
int	terrainSize = 32;

//////////////////////////////////////////////
//Global OpenGL rendering context variables
//////////////////////////////////////////////
GLuint vDefProg,fDefProg,pDef; //default shader prog
GLuint vPhongProg, fPhongProg, pPhong;
GLenum  ePolygonMode = GL_FILL;

///////////////////////////////////////////
//Program operation specific variables
///////////////////////////////////////////
int _time;
int lastTime;
int delta;

////////////////////////////////////////
//Main scene function prototypes
////////////////////////////////////////

//shader control function prototypes
int PrintOglError(char *file, int line);
void PrintShaderInfoLog(GLuint obj);
void PrintProgramInfoLog(GLuint obj);
void SetupDefShader();
void SetupPhongShader();

//Scene control function prototypes
void SetupSceneLighting();
void DrawScene();
void IdleScene();
void ResizeWindow(GLsizei iWidth, GLsizei iHeight);
void KeyboardInput(unsigned char cKey, int iX, int iY);
void DebugInfo();
void DrawString( char *text, float x, float y );
void CalcFPS(int delta_t);
void DrawInfo();

void DebugInfo()
{
	system("CLS");

	//stream any debug info here
}

void DrawString( char *text, float x, float y )
{
	char *c;
	int font=(int)GLUT_BITMAP_8_BY_13;

	glDisable( GL_LIGHTING );
	glDisable( GL_DEPTH_TEST );

	glColor3f( 1.0, 1.0, 1.0 );
	
	glMatrixMode(GL_PROJECTION);

	glPushMatrix();

	glLoadIdentity();
	gluOrtho2D( 0, 1024, 0, 768);

	glScalef(1, -1, 1);
	glTranslatef(0, -768, 0);
	glMatrixMode(GL_MODELVIEW);
	
	glPushMatrix();
	glLoadIdentity();
	
	glRasterPos2f( x, y );
	
	for( c=text; *c != '\0'; c++)
	{
		glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, *c);
	}
	
	glPopMatrix();
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	
	
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_LIGHTING );
}


void CalcFPS(int delta_t)
{
	fps_start = glutGet(GLUT_ELAPSED_TIME);

	fps_frames++;
	int fps = 0;

	fps = 1000.0 * fps_frames / delta_t;

	itoa(fps, fpsBuffer, 10);

    fps_frames = 0;
    fps_start = glutGet(GLUT_ELAPSED_TIME);
}

int main(int argc, char **argv) 
{ 
	///////////////////////////////////////
	//setup any instanced objects here
	///////////////////////////////////////
	gCamera = new Camera();

	GLsizei iWidth = 1024.0; 
    GLsizei iHeight = 768.0; 

    glutInit(&argc, argv);
	glutInitContextVersion(2, 0);
	//glutInitContextFlags(GLUT_FORWARD_COMPATIBLE);
	//glutInitContextProfile(GLUT_CORE_PROFILE);
    glutInitWindowPosition( 0, 0);
    glutInitWindowSize(iWidth, iHeight);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutCreateWindow( "OpenGL Voxel Terrain Demo" );
    glutDisplayFunc(DrawScene);
    glutIdleFunc(IdleScene);
    glutReshapeFunc(ResizeWindow);
    glutKeyboardFunc(KeyboardInput);

	//setup the camera
	glutSetCursor(GLUT_CURSOR_NONE); 
	gCamera->PositionCamera(50.0f, 120.0f, 250.0f,   0, 0, -1.0f,   0, 1, 0);

	//setup any openGL states here
	glEnable(GL_DEPTH_TEST);
	glClearColor (0.1, 0.57, 0.8, 1.0);
	glClearDepth(1.0f);
	glEnable(GL_CULL_FACE); //does this even have any effect?
	glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_SMOOTH);										
	glDepthFunc(GL_LEQUAL);		
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	
	glPolygonMode(GL_FRONT_AND_BACK, ePolygonMode);

	//Blending
	glBlendFunc(GL_SRC_COLOR, GL_ONE_MINUS_DST_COLOR);				// Set The Blending Function For Translucency
	glColor4f(0.0f, 1.0f, 0.0f, 0.4);

	glewInit();
	if (glewIsSupported("GL_VERSION_2_0"))
	{
		cout << "OpenGL 2.0 supported" << endl;
	}
	else 
	{
		cout << "OpenGL 2.0 not supported\n" << endl;
		exit(1);
	}

	gChunk.BuildCube();
	gChunk.InitVBO();

	SetupPhongShader();
	SetupDefShader();

	ResizeWindow(iWidth, iHeight);

	glutMainLoop();

	return 0;
}

//////////////////////////////////////////////////////////
//Shader control function prototype implementations
//////////////////////////////////////////////////////////
int PrintOglError(char *file, int line)
{
    GLenum glErr;
    int    retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)
    {
        printf("glError in file %s @ line %d: %s\n", file, line, gluErrorString(glErr));
        retCode = 1;
        glErr = glGetError();
    }
    return retCode;
}

void PrintShaderInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

	glGetShaderiv(obj, GL_INFO_LOG_LENGTH,&infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}

void PrintProgramInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

	glGetProgramiv(obj, GL_INFO_LOG_LENGTH,&infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *)malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}

/////////////////////////
//Setup shaders here
/////////////////////////
void SetupDefShader()
{
	char *vs = NULL,*fs = NULL;

	vDefProg = glCreateShader(GL_VERTEX_SHADER);
	fDefProg = glCreateShader(GL_FRAGMENT_SHADER);

	vs = textFileRead("data/minimalVert.c");
	fs = textFileRead("data/minimalFrag.c");

	const char * vv = vs;
	const char * ff = fs;

	glShaderSource(vDefProg, 1, &vv,NULL);
	glShaderSource(fDefProg, 1, &ff,NULL);

	free(vs);free(fs);

	glCompileShader(vDefProg);
	glCompileShader(fDefProg);

	PrintShaderInfoLog(vDefProg);
	PrintShaderInfoLog(fDefProg);

	pDef = glCreateProgram();
	glAttachShader(pDef,vDefProg);
	glAttachShader(pDef,fDefProg);

	glLinkProgram(pDef);
	PrintProgramInfoLog(pDef);

	cout << "Def shader loaded." << endl;
}

void SetupPhongShader()
{
	char *vs = NULL,*fs = NULL;

	//now the the same for the Phong shader
	vPhongProg = glCreateShader(GL_VERTEX_SHADER);
	fPhongProg = glCreateShader(GL_FRAGMENT_SHADER);

	vs = textFileRead("data/PhongVert.c");
	fs = textFileRead("data/PhongFrag.c");

	const char * vv = vs;
	const char * ff = fs;

	glShaderSource(vPhongProg, 1, &vv,NULL);
	glShaderSource(fPhongProg, 1, &ff,NULL);

	free(vs);free(fs);

	glCompileShader(vPhongProg);
	glCompileShader(fPhongProg);

	PrintShaderInfoLog(vPhongProg);
	PrintShaderInfoLog(fPhongProg);

	pPhong = glCreateProgram();
	glAttachShader(pPhong,vPhongProg);
	glAttachShader(pPhong,fPhongProg);

	glLinkProgram(pPhong);
	PrintProgramInfoLog(pPhong);
	
	cout << "Phong shader loaded." << endl;
}

////////////////////////////
//Resize window function
////////////////////////////
void ResizeWindow(GLsizei iWidth, GLsizei iHeight)
{
	glViewport( 0, 0, iWidth, iHeight ); 
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();

	//calculate aspect ratio
	gluPerspective(45.0f, iWidth/ iHeight, 1.0f, 500.0f);

	glMatrixMode(GL_MODELVIEW);// Select The Modelview Matrix
	glLoadIdentity();// Reset The Modelview Matrix
}

////////////////////////
//Main update loop
////////////////////////
void IdleScene()
{
	_time = glutGet(GLUT_ELAPSED_TIME);
	delta  = _time - lastTime;

	///////////////////////////////////
	//do any critical updates here
	///////////////////////////////////
	{
		//calculate the fps here
		CalcFPS(delta);
	}

	lastTime = _time;

	glutPostRedisplay();
}


//Draw text info here
void DrawInfo()
{
	DrawString("FPS: ", 5, 25);
	DrawString(fpsBuffer, 55, 25);

	if(useShaders)
	{
		DrawString("Using GLSL 1.1", 5, 45);
	}
	else
	{
		DrawString("Using emulated fixed function pipeline", 5, 45);
	}

	DrawString("Terrain voxel mesh size: ", 5, 65);
	itoa(terrainSize, sizeBuffer, 10);
	DrawString(sizeBuffer, 205, 65);
	DrawString("x", 235, 65);
	DrawString(sizeBuffer, 255, 65);
	DrawString("x", 285, 65);
	DrawString(sizeBuffer, 305, 65);
}

////////////////////////////////////////////////////////////////
//Setup global scene lighting parameters here
////////////////////////////////////////////////////////////////
void SetupSceneLighting()
{
	GLfloat Light_Position[] = {0.0f, 1000.0f, 0, 1.0f};
	GLfloat afPropertiesAmbient [] = {0.5, 0.8, 0.5, 1.00}; 
	GLfloat afPropertiesDiffuse [] = {0.5, 0.8, 0.5, 1.00}; 
	GLfloat afPropertiesSpecular[] = {0.7, 0.1, 0.7, 1.00}; 

	glEnable(GL_LIGHTING);
	glPolygonMode(GL_FRONT, ePolygonMode);

	glLightfv( GL_LIGHT0, GL_AMBIENT,  afPropertiesAmbient); 
	glLightfv( GL_LIGHT0, GL_DIFFUSE,  afPropertiesDiffuse); 
	glLightfv( GL_LIGHT0, GL_SPECULAR, afPropertiesSpecular); 
	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 0); 

	glEnable( GL_LIGHT0 ); 
}


//////////////////////
//Draw the scene
//////////////////////
void DrawScene()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ); 

	glLoadIdentity();

	//Update camera here
	gCamera->Update();

	////Draw scene geometry here
	glPushMatrix();
	{
		SetupSceneLighting();

		glPushMatrix();

		//draw scene data here
		if(useShaders)
		{
			glUseProgram(pPhong);
		}
		else
		{
			glUseProgramObjectARB(0);
		}

		gChunk.RenderVBOTerrain(1.0f, 1.0f, 1.0f, 32, gCamera);
		
		glPopMatrix();
	}

	DrawInfo();

	glPopMatrix();

	glutSwapBuffers();
}

/////////////////////////////
//Keyboard input function
/////////////////////////////
void KeyboardInput(unsigned char cKey, int iX, int iY)
{
	switch(cKey)
	{
	case '1':
		gChunk.ReInit(128);
		terrainSize = 128;
		lerpTerrain = true;
		break;

	case '2':
		gChunk.ReInit(64);
		terrainSize = 64;
		lerpTerrain = true;
		break;

	case '3':
		gChunk.ReInit(48);
		terrainSize = 48;
		lerpTerrain = true;
		break;

	case '4':
		gChunk.ReInit(32);
		terrainSize = 32;
		lerpTerrain = true;
		break;

	case '5':
		gChunk.ReInit(16);
		terrainSize = 16;
		lerpTerrain = true;
		break;

	case '6':
		gChunk.ReInit(4);
		terrainSize = 4;
		lerpTerrain = true;
		break;


	case 'f':
		useShaders = true;
		break;

	case 'g':
		useShaders = false;
		break;

	case 27: //Esc key
		exit(0);
		break;
	}
}