//Vector3 class implementation

#include "vector3.h"

Vector3 Vector3::Cross(Vector3 vVector1, Vector3 vVector2)
{
	Vector3 vNormal;	

	//Calculate the cross product
	vNormal.x = ((vVector1.y * vVector2.z) - (vVector1.z * vVector2.y));
	vNormal.y = ((vVector1.z * vVector2.x) - (vVector1.x * vVector2.z));
	vNormal.z = ((vVector1.x * vVector2.y) - (vVector1.y * vVector2.x));

	return vNormal;		
}

GLfloat Vector3::Magnitude(Vector3 vNormal)
{
	//Calculate the magnitude of the vector
	return (GLfloat)sqrt( (vNormal.x * vNormal.x) + 
						(vNormal.y * vNormal.y) + 
						(vNormal.z * vNormal.z) );
}

Vector3 Vector3::Normalize(Vector3 vVector)
{
	GLfloat magnitude = Magnitude(vVector);				

	//Divide the original vector by the magnitude vector to find the normal vector  
	vVector = vVector / magnitude;		

	return vVector;		
}

Vector3 Vector3::Vector(Vector3 vPoint1, Vector3 vPoint2)
{
	Vector3 vVector = Vector3(0, 0, 0);
	vVector.x = vPoint1.x - vPoint2.x;					// Get the X value of our new vector
	vVector.y = vPoint1.y - vPoint2.y;					// Get the Y value of our new vector
	vVector.z = vPoint1.z - vPoint2.z;					// Get the Z value of our new vector

	return vVector;				
}

Vector3 Vector3::Normal(Vector3 vTriangle[])					
{														// Get 2 vectors from the polygon (2 sides), Remember the order!
	Vector3 vVector1 = Vector(vTriangle[2], vTriangle[0]);
	Vector3 vVector2 = Vector(vTriangle[1], vTriangle[0]);

	Vector3 vNormal = Cross(vVector1, vVector2);		// Take the cross product of our 2 vectors to get a perpendicular vector

	// Now we have a normal, but it's at a strange length, so let's make it length 1.

	vNormal = Normalize(vNormal);						// Use our function we created to normalize the normal (Makes it a length of one)

	return vNormal;										// Return our normal at our desired length
}

float Vector3::Dot(Vector3 vVector1, Vector3 vVector2)
{
	return ((vVector1.x * vVector2.x) + (vVector1.y * vVector2.y) + (vVector1.z * vVector2.z));
}

float Vector3::AngleBetweenVectors(Vector3 Vector1, Vector3 Vector2)
{
	float dotProduct = Dot(Vector1, Vector2);				

	float vectorsMagnitude = Magnitude(Vector1) * Magnitude(Vector2) ;

	double angle = acos( dotProduct / vectorsMagnitude );

	// Return the angle in radians
	return( angle );
}